/**
  *  \file planetscentral/dbexport/dbexport.cc
  *  \brief Database Exporter: Main
  */

#include <iostream>
#include <cstring>
#include <stdexcept>
#include "planetscentral/dbexport/dbexport.h"
#include "planetscentral/osutil.h"
#include "cpluslib/string.h"
#include "planetscentral/dbclient/dbclient.h"
#include "server/config.h"
#include "int/args.h"
#include "int/data.h"
#include "planetscentral/dbexport/exdb.h"

/******************** Configuration and Global Objects *******************/

DbReadyClient database_connection;
string_t database_host = "127.0.0.1";
string_t database_port = "6379";

static bool
checkConfig1(string_t key, string_t value)
{
    if (key == "REDIS.HOST") {
        database_host = value;
        return true;
    } else if (key == "REDIS.PORT") {
        database_port = value;
        return true;
    } else {
        return false;
    }
}

static void
checkConfig(string_t key, string_t value)
{
    checkConfig1(key, value);
}

/*************************** Command Dispatcher **************************/

static void
executeCommand(const char*const* argv)
{
    // Convert into usable format
    const char* cmd = *argv;
    IntDataSegment dseg;
    while (const char* p = *++argv) {
        dseg.pushNew(makeStringValue(p));
    }
    IntArgBlock args(dseg, 0, dseg.size());

    // Dispatch
    if (std::strcmp(cmd, "db") == 0) {
        database_connection.connectRetry(database_host, database_port);
        doDatabaseExport(std::cout, args);
    } else {
        throw std::runtime_error("Unsupported command");
    }
}

/********************************** Main *********************************/

int main(int /*argc*/, char** argv)
{
    // Read config and parse options
    loadConfiguration(checkConfig, true);
    while (const char* p = *++argv) {
        if (*p == '-') {
            // "--foo" and "-foo" are the same
            if (p[1] == '-') {
                ++p;
            }

            // Help
            if (strcmp(p, "-help") == 0) {
                std::cout << "c2dbexport - PlanetsCentral Database Export\n\n"
                          << "Usage: c2dbexport [--OPTION=VALUE] COMMAND [ARGS...]\n\n"
                          << "--OPTION=VALUE   overrides config file settings\n\n"
                          << "Commands:\n"
                          << "  db [--delete] WILDCARD...     export database keys\n"
                          << "\nThis utility creates c2console (*.con) scripts to restore\n"
                          << "a particular situation / set of data in the same or another\n"
                          << "PlanetsCentral database instance.\n";
                return 0;
            } else {
                const char* q = strchr(p, '=');
                if (q != 0 && checkConfig1(strUCase(string_t(p+1, q-p-1)), q+1)) {
                    // ok
                } else {
                    std::cerr << "c2dbexport: unknown option '" << p << "'\n";
                }
            }
        } else {
            break;
        }
    }

    // Do we have a command?
    if (*argv == 0) {
        std::cerr << "c2dbexport: no command specified\n";
        return 1;
    }

    // Execute command
    try {
        executeCommand(argv);
        return 0;
    }
    catch (std::exception& e) {
        std::cerr << "c2dbexport: " << e.what() << "\n";
        return 1;
    }
}
