/**
  *  \file planetscentral/host/cmdgame.cc
  *  \brief Host: GAME commands
  */

#include "planetscentral/host/cmdgame.h"
#include "cpluslib/format.h"
#include "game/struct.h"
#include "int/args.h"
#include "int/array.h"
#include "planetscentral/dbclient/dbsubtree.h"
#include "planetscentral/host/db.h"
#include "planetscentral/host/host.h"
#include "planetscentral/host/lock.h"
#include "planetscentral/osutil.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/host/game.h"
#include "planetscentral/host/cmdhost.h"
#include "planetscentral/dbclient/fileclient.h"
#include "planetscentral/host/victory.h"
#include "planetscentral/host/cron.h"
#include "planetscentral/host/newgame.h"
#include "planetscentral/host/talk.h"

/* @q NEWGAME (Host Command)
   Create new game.
   The game is created in status "preparing", type "private".
   Use other commands to configure it.

   Permissions: ?

   @retval GID new game Id
   @uses game:state:$STATE, game:all, game:lastid, game:$GID:name, game:$GID:settings */
void
doNewGame(HostConnection& conn, IntArgBlock& /*args*/)
{
    // NEWGAME
    DbSubtree root(database_connection, GAME_ROOT);

    int32_t id = createNewGame();
    {
        // Prepare it
        ReconnectingGuard g(database_rc);
        DbSubtree game(root.subtree(id));
        game.stringKey("name").set("New Game");
        game.stringKey("owner").set("");
        game.intKey("schedule:lastId").set(0);
        game.hashKey("settings").setField("description", "New Game");
        game.hashKey("settings").setField("host",   DbSubtree(database_connection, HOST_PROGRAM_ROOT).stringKey("default").get());
        game.hashKey("settings").setField("master", DbSubtree(database_connection, MASTER_PROGRAM_ROOT).stringKey("default").get());
        game.hashKey("settings").setField("shiplist", DbSubtree(database_connection, SHIPLIST_PROGRAM_ROOT).stringKey("default").get());

        // Player slots
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            DbHash h = game.subtree("player").subtree(i).hashKey("status");
            h.setField("slot", 1);  // Slot is open
            h.setField("turn", 0);  // Turn is missing
        }
    }

    // Place game on lists. Only after this, the game will be
    // visible to other commands.
    createNewGameFinish(id, "preparing", "private");

    IntIntIntValue iv(id);
    conn.sendObject(&iv);
}

/* @q CLONEGAME src:GID [state:HostGameState] (Host Command)
   Clone a game.
   The new game is created in state "joining" if no initial state is given.

   Permissions: admin for game.

   @retval GID new game Id */
void
doCloneGame(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1, 2);
    int32_t gameId = toIntResult(args.getNext());
    string_t state = args.getNumArgs() > 0 ? toStringResult(args.getNext()) : "joining";
    if (state != "preparing" && state != "joining" && state != "running" && state != "finished" && state != "deleted") {
        conn.sendError("400 Invalid game state");
        return;
    }

    // Lock source game
    GameLock gl(gameId);
    gl.lockOrDie(IntentDB);

    // Check that source exists, and get type
    string_t gameType;
    {
        ReconnectingGuard g(database_rc);
        Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
        conn.checkPermission(game, Game::AdminPermission);
        gameType = game.getType();
    }

    // Do it
    int32_t id = createNewGame();
    {
        ReconnectingGuard g(database_rc);
        copyGame(gameId, id);
    }
    createNewGameFinish(id, state, gameType);

    // Result
    IntIntIntValue iv(id);
    conn.sendObject(&iv);
}

/* @q GAMESETTYPE game:GID type:HostGameType (Host Command)
   Set game type.

   Permissions: admin for game.

   @uses game:$GID:type, game:pubstate:$STATE */
void
doSetGameType(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(2);
    int32_t gameId = toIntResult(args.getNext());
    string_t type  = toStringResult(args.getNext());
    if (type != "private" && type != "unlisted" && type != "public" && type != "test") {
        conn.sendError("400 Invalid game type");
        return;
    }

    // Lock game
    GameLock gl(gameId);
    gl.lockOrDie(IntentDB);

    // Execute
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::AdminPermission);
    game.setType(type);
    conn.sendSuccess("OK");
}

/* @q GAMESETSTATE game:GID type:HostGameState (Host Command)
   Set game state.
   This will possibly start or stop the scheduler.

   Permissions: admin for game.

   @uses game:pubstate:$STATE, game:state:$STATE, game:$GID:state */
void
doSetGameState(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(2);
    int32_t gameId = toIntResult(args.getNext());
    string_t state = toStringResult(args.getNext());
    if (state != "preparing" && state != "joining" && state != "running" && state != "finished" && state != "deleted") {
        conn.sendError("400 Invalid game state");
        return;
    }

    // Lock game
    GameLock gl(gameId);
    gl.lockOrDie(IntentDB);

    // Execute
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::AdminPermission);
    if (state == "joining") {
        // reset change flags
        game.configChanged().remove();
        game.endChanged().remove();
        game.scheduleChanged().remove();
    }
    if (state == "finished") {
        // do ranks first so setState sees the ranks for generating history
        checkForcedGameEnd(game);
        computeGameRankings(database_connection, game);
    }
    game.setState(state);
    conn.sendSuccess("OK");
}

/* @q GAMESETOWNER game:GID user:UID (Host Command)
   Set game owner.

   Permissions: admin for game.

   @uses game:$GID:owner, user:$UID:ownedGames */
void
doSetGameOwner(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(2);
    int32_t gameId = toIntResult(args.getNext());
    string_t user  = toStringResult(args.getNext());

    // Lock game. Just to be sure.
    GameLock gl(gameId);
    gl.lockOrDie(IntentDB);

    // Change owner.
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::AdminPermission);
    game.setOwner(user, DbSubtree(database_connection, USER_ROOT));
    conn.sendSuccess("OK");
}

/* @q GAMESETNAME game:GID name:Str (Host Command)
   Set game name.

   Permissions: admin for game.

   @uses game:$GID:name */
void
doSetGameName(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(2);
    int32_t gameId = toIntResult(args.getNext());
    string_t name  = toStringResult(args.getNext());

    // Change name. No more synchronisation required because the name is pretty uncritical.
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::AdminPermission);
    game.setName(name);
    conn.sendSuccess("OK");
}

/* @q GAMESTAT game:GID (Host Command)
   Get information about one game.

   Permissions: read-access to game.

   @retval HostGameInfo Verbose information */
void
doStatGame(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    int32_t gameId = toIntResult(args.getNext());

    // Database stuff
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::ReadPermission);
    std::auto_ptr<IntValue> result(game.describe(true, conn.getUser()));
    conn.sendObject(result.get());
}

/* @q GAMELIST [STATE state:HostGameState] [TYPE type:HostGameType] [USER user:UID] [VERBOSE|ID] (Host Command)
   List games.
   Optional parameters %state, %type, %user limit output to games of that state, type, or played by that user.

   The command returns a list of {@type HostGameInfo} (normal) by default.
   %VERBOSE returns the verbose variant, %ID returns just a list of {@type GID|game Ids}.

   Permissions: returns only games for which user has read access.

   @rettype HostGameInfo
   @rettype IntList
   @rettype GID
   @uses game:state:$STATE, game:pubstate:$STATE, game:all, user:$UID:ownedGames */
void
doListGames(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    string_t stateLimit;
    string_t typeLimit;
    string_t forUser;       // FIXME
    // FIXME: permissions
    enum { Normal, Verbose, Ids } mode = Normal;
    while (args.getNumArgs() > 0) {
        string_t keyword = strUCase(toStringResult(args.getNext()));
        if (keyword == "STATE") {
            args.checkArgumentCountAtLeast(1);
            stateLimit = toStringResult(args.getNext());
        } else if (keyword == "TYPE") {
            args.checkArgumentCountAtLeast(1);
            typeLimit = toStringResult(args.getNext());
        } else if (keyword == "USER") {
            args.checkArgumentCountAtLeast(1);
            forUser = toStringResult(args.getNext());
        } else if (keyword == "VERBOSE") {
            mode = Verbose;
        } else if (keyword == "ID") {
            mode = Ids;
        } else {
            conn.sendError("400 Syntax error");
            return;
        }
    }

    // Permission checks
    bool needPermissionCheck;
    bool needTypeCheck;
    bool needStateCheck;
    if (!conn.isAdmin()) {
        needPermissionCheck = (stateLimit != "joining" && stateLimit != "running" && stateLimit != "finished")
            || (typeLimit != "public" || typeLimit != "unlisted");
    } else {
        needPermissionCheck = false;
    }

    // Work
    ReconnectingGuard g(database_rc);
    DbSubtree root(database_connection, GAME_ROOT);
    IntList_t games;

    if (!forUser.empty()) {
        // User's games.
        // Permission check will wield out deleted/preparing games.
        IntList_t gameRefs;
        DbSubtree(database_connection, USER_ROOT).subtree(forUser).hashKey("games").getAll(gameRefs);
        needTypeCheck = true;
        needStateCheck = true;
        for (IntList_t::size_type i = 0; i < gameRefs.size(); i += 2) {
            if (gameRefs[i+1] != 0) {
                games.push_back(gameRefs[i]);
            }
        }
    } else if (stateLimit.empty()) {
        // No state limit given
        root.intSetKey("all").getAll(games);
        needTypeCheck = true;
        needStateCheck = false;
    } else {
        // State limit given, so use by-state lists
        needStateCheck = false;
        if (typeLimit == "public") {
            root.subtree("pubstate").intSetKey(stateLimit).getAll(games);
            needTypeCheck = false;
        } else {
            root.subtree("state").intSetKey(stateLimit).getAll(games);
            needTypeCheck = true;
        }
    }
    if (typeLimit.empty()) {
        needTypeCheck = false;
    }
    if (stateLimit.empty()) {
        needStateCheck = false;
    }

    Ptr<IntArrayData> a = new IntArrayData();
    for (uint32_t i = 0; i < games.size(); ++i) {
        Game game(root, games[i], Game::NoExistanceCheck);
        if ((!needPermissionCheck || game.hasPermission(conn.getUser(), Game::ReadPermission))
            && (!needTypeCheck || game.getType() == typeLimit)
            && (!needStateCheck || game.getState() == stateLimit))
        {
            if (mode == Ids) {
                a->content.pushNew(makeIntValue(games[i]));
            } else {
                a->content.pushNew(game.describe(mode == Verbose, conn.getUser()));
            }
        }
    }
    a->addDimension(a->content.size());

    IntArray result(a);
    conn.sendObject(&result);
}

/* @q GAMESET game:GID [key:Str value:Str ...] (Host Command)
   Set game properties.
   Any number of properties can be set at once by listing multiple key/value pairs.
   See {game:$GID:settings} for possible keys.

   Permissions: config-access to game.

   @err 412 Value not permitted (the specified value is not valid for this key)

   @uses game:$GID:settings */
void
doSetGameConfig(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(1);
    int32_t gameId = toIntResult(args.getNext());

    // Operate
    GameLock gl(gameId);
    gl.lockOrDie(IntentDB);

    // Validate and build command
    ReconnectingGuard g(database_rc);
    StringList_t cmd;
    bool cronChange = false;
    bool configChanged = false;
    bool endChanged = false;
    bool configSet = false;
    bool endSet = false;
    while (args.getNumArgs() > 0) {
        args.checkArgumentCountAtLeast(2);
        string_t option = toStringResult(args.getNext());
        string_t value = toStringResult(args.getNext());

        // Verify options
        if (option == "host" && !DbSubtree(database_connection, HOST_PROGRAM_ROOT).stringSetKey("list").contains(value)) {
            conn.sendError("412 Value not permitted");
            return;
        }
        if (option == "master" && !DbSubtree(database_connection, MASTER_PROGRAM_ROOT).stringSetKey("list").contains(value)) {
            conn.sendError("412 Value not permitted");
            return;
        }
        if (option == "shiplist" && !DbSubtree(database_connection, SHIPLIST_PROGRAM_ROOT).stringSetKey("list").contains(value)) {
            conn.sendError("412 Value not permitted");
            return;
        }

        // Options that trigger special behaviour
        if (option == "hostRunNow") {
            cronChange = true;
        }
        if (option == "host" || option == "master" || option == "shiplist") {
            configChanged = true;
        }
        if (option.length() >= 3 && option.compare(0, 3, "end", 3) == 0) {
            endChanged = true;
        }
        if (option == "endChanged") {
            endSet = true;
        }
        if (option == "configChanged") {
            configSet = true;
        }
        cmd.push_back(option);
        cmd.push_back(value);
    }

    // Execute
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::ConfigPermission);
    game.setConfig(cmd);
    game.clearCache();
    conn.sendSuccess("OK");

    // Set status bits
    if (endChanged && !endSet) {
        game.endChanged().set(1);
    }
    if (configChanged && !configSet) {
        game.configChanged().set(1);
    }

    // Postprocess
    if (cronChange) {
        cron.handleGameChange(gameId);
    }
}

/* @q GAMEGET game:GID key:Str (Host Command)
   Get game property.
   See {game:$GID:settings} for possible keys.

   Permissions: read-access to game.

   @retval Str property
   @uses game:$GID:settings
   @see GAMESET, GAMEMGET */
void
doGetGameConfig(HostConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCount(2);
    int32_t gameId = toIntResult(args.getNext());
    string_t option = toStringResult(args.getNext());

    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::ReadPermission);
    IntStringValue sv(game.getConfig(option));
    conn.sendObject(&sv);
}

/* @q GAMEMGET game:GID key:Str... (Host Command)
   Get game properties.
   Like {GAMEGET}, but returns multiple values at once as an array.

   Permissions: read-access to game.

   @retval StrList values
   @uses game:$GID:settings
   @see GAMESET, GAMEGET */
void
doGetGameConfigMultiple(HostConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCountAtLeast(1);
    int32_t gameId = toIntResult(args.getNext());
    StringList_t fields;
    while (args.getNumArgs() > 0) {
        fields.push_back(toStringResult(args.getNext()));
    }

    // Do it
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::ReadPermission);
    std::auto_ptr<IntValue> result(game.getConfig(fields));
    conn.sendObject(result.get());
}

/* @q GAMEGETCC game:GID key:Str (Host Command)
   Get computed/cached value.
   The value is fetched from the cache; if it is not yet cached, it is computed
   Value keys:
   - %difficulty

   Permissions: read-access to game.

   @uses game:$GID:cache */
void
doGetGameCache(HostConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCountAtLeast(2);
    int32_t gameId = toIntResult(args.getNext());
    string_t option = toStringResult(args.getNext());

    // Permission checks
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::ReadPermission);

    // Do it
    std::auto_ptr<IntValue> result;
    if (option == "difficulty") {
        IntIntIntValue iv(game.getDifficulty());
        conn.sendObject(&iv);
    } else {
        conn.sendError("404 Unknown value");
    }
}

/* @q GAMEGETSTATE game:GID (Host Command)
   Get game state.

   Permissions: read-access to game.

   @uses game:$GID:state
   @retval HostGameState */

/* @q GAMEGETTYPE game:GID (Host Command)
   Get game type.

   Permissions: read-access to game.

   @uses game:$GID:type
   @retval HostGameType */

/* @q GAMEGETOWNER game:GID (Host Command)
   Get game owner.

   Permissions: read-access to game.

   @uses game:$GID:owner
   @retval UID */

/* @q GAMEGETNAME game:GID (Host Command)
   Get game name.

   Permissions: read-access to game.

   @uses game:$GID:name
   @retval Str */

/* @q GAMEGETDIR game:GID (Host Command)
   Get game directory in host filer.

   Permissions: read-access to game.

   @uses game:$GID:dir
   @retval FileName */
void
doGetGameProperty(HostConnection& conn, IntArgBlock& args, string_t prop)
{
    args.checkArgumentCount(1);
    int32_t gameId = toIntResult(args.getNext());

    ReconnectingGuard g(database_rc);
    DbSubtree root(database_connection, GAME_ROOT);
    Game game(root, gameId);
    conn.checkPermission(game, Game::ReadPermission);

    // This bypasses the getState(), getOwner(), etc. methods. Probably fix it someday.
    IntStringValue sv(root.subtree(gameId).stringKey(prop).get());
    conn.sendObject(&sv);
}

/* @q GAMECHECKPERM game:GID (Host Command)
   Get game access permissions.

   Permissions: none.

   @err 404 Game does not exist
   @retval Int Sum of permissions the current user has on this game
   - 1 (user is owner)
   - 2 (user is primary player in this game)
   - 4 (user is active (in-charge) replacement player in this game)
   - 8 (user is inactive replacement in this game)
   - 16 (game is public) */
void
doCheckGamePermission(HostConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCount(2);
    int32_t gameId = toIntResult(args.getNext());
    string_t userId = toStringResult(args.getNext());

    ReconnectingGuard g(database_rc);
    DbSubtree root(database_connection, GAME_ROOT);
    if (!root.intSetKey("all").contains(gameId)) {
        conn.sendError("404 Game does not exist");
        return;
    }

    int32_t value = 0;

    DbSubtree game(root.subtree(gameId));

    string_t state = game.stringKey("state").get();
    if (state != "deleted" && state != "preparing") {
        if (game.stringKey("owner").get() == userId) {
            // 1 - owner
            value |= 1;
        }

        // Player checks
        for (int i = 1; i <= 11; ++i) {
            StringList_t list;
            game.subtree("player").subtree(i).stringListKey("users").getAll(list);
            if (!list.empty()) {
                if (userId == list.back()) {
                    // 2 - primary
                    value |= 2;
                }
                if (userId == list.front()) {
                    // 4 - current
                    value |= 4;
                }
                for (uint32_t i = 1; i+1 < list.size(); ++i) {
                    if (userId == list[i]) {
                        // 8 - inactive replacement
                        value |= 8;
                        break;
                    }
                }
            }
        }

        if (value == 0 && game.stringKey("type").get() == "public") {
            // You're a member of the general public
            value |= 16;
        }
    }

    IntIntIntValue iv(value);
    conn.sendObject(&iv);
}

/* @q GAMEADDTOOL game:GID tool:Str (Host Command)
   Add a tool.
   If another tool of the same type already is on the game, remove that first.

   Permissions: admin for game.

   @err 404 Invalid tool
   @uses game:$GID:tools, game:$GID:toolkind, game:$GID:tool:$TOOL:settings
   @retval Int 1=tool changed, 0=no change */

/* @q GAMERMTOOL game:GID tool:Str (Host Command)
   Add a tool.

   Permissions: admin for game.

   @err 404 Invalid tool
   @uses game:$GID:tools, game:$GID:toolkind, game:$GID:tool:$TOOL:settings
   @retval Int 1=tool changed, 0=no change */
void
doChangeGameTool(HostConnection& conn, IntArgBlock& args, bool add)
{
    // Parse args
    args.checkArgumentCount(2);
    int32_t gameId  = toIntResult(args.getNext());
    string_t toolId = toStringResult(args.getNext());

    // Operate
    GameLock gl(gameId);
    gl.lockOrDie(IntentDB);

    // Validate
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::AdminPermission);
    if (!DbSubtree(database_connection, TOOL_PROGRAM_ROOT).stringSetKey("list").contains(toolId)) {
        conn.sendError("404 Invalid tool");
        return;
    }

    // Do it
    DbStringSet tools = game.getGame().stringSetKey("tools");
    bool result = add ? tools.add(toolId) : tools.remove(toolId);

    // If this was a change, update tool kinds
    if (result) {
        DbHash tool(DbSubtree(database_connection, TOOL_PROGRAM_ROOT).subtree("prog").hashKey(toolId));
        string_t kind = tool.getStringField("kind");
        if (kind.empty()) {
            // Tool has no kind, no conflict check
        } else if (add) {
            // We have added a tool. Check for another tool of the same kind.
            string_t oldTool = game.getGame().hashKey("toolkind").getStringField(kind);
            if (oldTool.size() && oldTool != toolId) {
                tools.remove(oldTool);
                game.getGame().subtree(oldTool).hashKey("settings").remove();
            }
            game.getGame().hashKey("toolkind").setField(kind, toolId);
        } else {
            // We have removed a tool.
            game.getGame().hashKey("toolkind").removeField(kind);
            game.getGame().subtree(toolId).hashKey("settings").remove();
        }
        game.clearCache();
        game.configChanged().set(1);
    }

    IntIntIntValue iv(result);
    conn.sendObject(&iv);
}

/* @q GAMELSTOOLS game:GID (Host Command)
   List game tools (addons).

   Permissions: read-access to game.

   @retval HostToolInfo[]
   @uses game:$GID:tools, prog:tool:prog:$TOOL */
void
doListGameTools(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    int32_t gameId = toIntResult(args.getNext());

    // Operate
    ReconnectingGuard g(database_rc);
    DbSubtree base(database_connection, TOOL_PROGRAM_ROOT);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::ReadPermission);

    StringList_t tools;
    game.getGame().stringSetKey("tools").getAll(tools);

    Ptr<IntArrayData> result = new IntArrayData;
    for (uint32_t i = 0, n = tools.size(); i < n; ++i) {
        result->content.pushNew(describeHost(base, tools[i], false));
    }
    result->addDimension(result->content.size());
    IntArray a(result);
    conn.sendObject(&a);
}

/* @q GAMETOTALS (Host Command)
   Get host statistics.

   Permissions: none.

   @retkey joining:Int Number of joinable public games
   @retkey running:Int Number of running public games
   @retkey finished:Int Number of finished public games
   @uses game:pubstate:$STATE */
void
doGameTotals(HostConnection& conn, IntArgBlock& args)
{
    // Args
    args.checkArgumentCount(0);

    // Operate
    ReconnectingGuard g(database_rc);
    DbSubtree pubstate(DbSubtree(database_connection, GAME_ROOT).subtree("pubstate"));

    Ptr<IntArrayData> result = new IntArrayData;
    result->content.pushNew(makeStringValue("joining"));
    result->content.pushNew(makeIntValue(pubstate.intSetKey("joining").size()));
    result->content.pushNew(makeStringValue("running"));
    result->content.pushNew(makeIntValue(pubstate.intSetKey("running").size()));
    result->content.pushNew(makeStringValue("finished"));
    result->content.pushNew(makeIntValue(pubstate.intSetKey("finished").size()));
    result->addDimension(result->content.size());
    IntArray a(result);
    conn.sendObject(&a);
}

/* @q GAMEGETVC game:GID (Host Command)
   Get victory condition.

   Permissions: read-access to game.

   @retkey endCondition:Str (end condition type, "turn", "score", or none)
   @retkey endTurn:Int (for "turn" condition, ending turn; for "score" condition, number of turns to hold a score)
   @retkey endProbability:Int (for "turn" condition, probability that game ends at that turn)
   @retkey endScore:Int (for "score" condition: score to reach and hold)
   @retkey endScoreName:Str (for "score" condition: name of score)
   @retkey endScoreDescription:Str (for "score" condition: description of score)
   @retkey referee:Str (for no condition: name of referee add-on)
   @retkey refereeDescription:Str (for no condition: description of referee add-on, see {prog:tool:prog:$TOOL}->description)
   @uses game:$GID:settings */
void
doGameGetVictoryCondition(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    int32_t gameId = toIntResult(args.getNext());

    // Operate
    ReconnectingGuard g(database_rc);
    DbSubtree base(database_connection, TOOL_PROGRAM_ROOT);
    DbSubtree root(database_connection, GAME_ROOT);
    Game game(root, gameId);
    conn.checkPermission(game, Game::ReadPermission);

    std::auto_ptr<IntValue> result(describeVictoryCondition(game, database_connection));
    conn.sendObject(result.get());
}

/* @q GAMEUPDATE game:GID... (Host Command)
   Ad-hoc, admin-only command to update a game to the latest data formats.

   Permissions: admin. */
void
doGameUpdate(HostConnection& conn, IntArgBlock& args)
{
    conn.checkAdmin();

    ReconnectingGuard g(database_rc);
    while (args.getNumArgs() > 0) {
        // Fetch a game
        int32_t gameId = toIntResult(args.getNext());
        DbSubtree root(database_connection, GAME_ROOT);
        Game game(root, gameId);

        // Get game specs
        string_t gameState = game.getState();
        string_t gameType  = game.getType();

        // Update this game's forum
        if (gameState == "joining" || gameState == "running") {
            talkHandleGameStart(game, gameType);
        }
    }
    conn.sendSuccess("OK");
}
