/**
  *  \file planetscentral/host/lock.cc
  *  \brief Host: Locking
  *
  *  This provides a means to "lock" a game against write operations.
  */

#include <map>
#include <stdexcept>
#include <unistd.h>        // usleep
#include "planetscentral/host/lock.h"
#include "planetscentral/osutil.h"

static std::map<int32_t, int32_t> lockedGames;
static Mutex lockedGamesMutex;

static LockIntent
tryLockInternal(int32_t gameId, LockIntent intent)
{
    MutexGuard g(lockedGamesMutex);
    std::map<int32_t, int32_t>::iterator it = lockedGames.find(gameId);
    if (it == lockedGames.end()) {
        lockedGames.insert(std::make_pair(gameId, int32_t(intent)));
        return NoIntent;
    } else {
        return LockIntent(it->second);
    }
}

/** Lock a game.
    \param gameId Game to lock
    \param intent Intent to lock the game for (must not be NoIntent)
    \retval true lock obtained
    \retval false lock cannot be obtained */
bool
lockGame(int32_t gameId, LockIntent intent)
{
    int32_t timer = 0;
    while (1) {
        /*           ------- oldIntent ----------
           intent    No    DB    Chktrn   Runhost
           --------  ----------------------------
           DB        OK    Wait  Wait     Fail
           Chktrn    OK    Wait  Wait     Fail
           Runhost   OK    Wait  Wait     (Fail)

           The idea is that database and checkturn accesses are fast,
           so we can wait for them. Runhost is slow, and accesses
           immediately fail.

           FIXME: this implements dumbass polling. Should this become
           a problem, we should implement real waitqueues. */
        LockIntent oldIntent = tryLockInternal(gameId, intent);
        if (oldIntent == NoIntent) {
            return true;
        }
        if (oldIntent == IntentRunHost) {
            return false;
        }
        if (++timer > 100) {
            return false;
        }
        usleep(50000);
    }
}

/** Unlock game. The game must have previously been locked by the same
    thread.
    \param gameId Game to unlock */
void
unlockGame(int32_t gameId)
{
    MutexGuard g(lockedGamesMutex);
    lockedGames.erase(gameId);
}

GameLock::GameLock(int32_t gameId)
    : isLocked(false),
      gameId(gameId)
{ }

GameLock::~GameLock()
{
    if (isLocked) {
        unlockGame(gameId);
    }
}

bool
GameLock::lock(LockIntent intent)
{
    if (!isLocked) {
        isLocked = lockGame(gameId, intent);
    }
    return isLocked;
}

void
GameLock::lockOrDie(LockIntent intent)
{
    if (!lock(intent)) {
        throw std::runtime_error("600 Game in use");
    }
}
